"""Example Future Strategy (No Exit)"""

# Strategy: 1m MACDEMA
# Long if price above 200 EMA and MACD crosses below 0
# Short if price below 200 EMA and MACD crosses above 0

from tradeboy.tools import Tools

class Strategy:
    def __init__(self):
        # Helper class
        self.tools = Tools(exchange='phemex', verbose=False)
        # Required params needed to initiate trade
        self.params = {
            'type': 'market',  # Or limit
            'side': None,  # long or short
            'amount': 1,  # How many contracts to trade with
            'tp_percent': None,  # Take profit percent
            'sl_percent': None,  # Stop loss percent
            'symbol': self.tools.symbol(base='BTC', quote='USD', code='future'), # Trading pair,
            'exit': False # Exit signal
        }

    def get_macd(self):
        # Get macd values
        # Use default params (12, 26, 9)
        _, _, _, _, close, _ = self.tools.ohlcv(symbol=self.params['symbol'], tf='1m')
        macd, signal, _ = self.tools.macd(close)
        # Format
        macd = list(macd)
        macd.reverse()
        signal = list(signal)
        signal.reverse()

        # Get current and previous values
        prev_macd = None
        last_macd = None
        prev_signal = None
        last_signal = None

        i = 0
        for value in macd:
            if str(value) != 'nan':
                prev_macd = macd[i + 1]
                last_macd = macd[i + 2]
                break
            i += 1

        i = 0
        for value in signal:
            if str(value) != 'nan':
                prev_signal = signal[i + 1]
                last_signal = signal[i + 2]
                break
            i += 1

        return prev_macd, last_macd, prev_signal, last_signal

    def get_ema(self):
        # Get ema value
        # Use default timeperiod = 200
        _, _, _, _, close, _ = self.tools.ohlcv(symbol=self.params['symbol'], tf='1m')
        ema = self.tools.ema(close)

        # Format
        ema = list(ema)
        ema.reverse()

        # Get current value
        curr_ema = None

        for value in ema:
            if str(value) != 'nan':
                curr_ema = value
                break

        return curr_ema

    def signal(self):
        # Get strategy params
        price = self.tools.price(symbol=self.params['symbol'])
        prev_macd, last_macd, prev_signal, last_signal = self.get_macd()
        curr_ema = self.get_ema()

        # Check for entry
        if price > curr_ema:
            # Look for long
            if last_macd < last_signal and prev_macd > prev_signal and prev_macd < 0 and prev_signal < 0:
                self.params['tp_percent'] = 0.36  # 0.36%
                self.params['sl_percent'] = 0.18  # 0.18%
                self.params['side'] = 'long'
                return True

        if price < curr_ema:
            # Look for short
            if last_macd > last_signal and prev_macd < prev_signal and prev_macd > 0 and prev_signal > 0:
                self.params['tp_percent'] = 0.36
                self.params['sl_percent'] = 0.18
                self.params['side'] = 'short'
                return True

        return False

    # Entry is required
    # Exit is not required if tp/sl percents provided
    def entry(self):
        # Returns True if signal was found
        # False if signal was not found
        return self.signal()

    # Exit conditions not needed for current strategy
    def exit(self):
        pass



"""General use functions"""

import datetime

from time import sleep

def sync():
    """Waits until the new minute starts"""
    t = datetime.datetime.utcnow()
    x = 60 - (t.second + t.microsecond/1000000.0)
    sleep(x)



"""Automated Crypto Trader for future market"""

from phemexboy.proxy import Proxy
from tradeboy.helpers.utility import sync
from tradeboy.helpers.calc import take_profit, stop_loss
from tradeboy.helpers.exceptions import InvalidParamError
from time import sleep

# TODO: Add cleanup method (make sure no position open)
# TODO: Exit strategy, limit
# TODO: Keep running after timeout error
# TODO: Send logs to discord
# TODO: trailing stop loss
# TODO: Tools error handling and tests

class FutureEngine:
  def __init__(self, Strategy: object, rate: int = 20, verbose: bool = False):
    usd_balance = 0
    try:
      self._strategy = Strategy()
      self._proxy = Proxy(verbose=verbose)
      usd_balance = self._proxy.balance(currency='USD', code='future')
    except Exception as e:
      print(f'Failed to initialize engine: \n{e}')
    finally:
      self._rate = rate # Request rate
      self._verbose = verbose
      self._trade_stats = {
        "trade": 1,
        "wins": 0,
        "losses": 0
      }
      self._position_stats = {
        "entry": 0,
        "tp": 0,
        "sl": 0
      }
      self._wallet = {
        "balance": usd_balance,
        "pnl": 0,
      }
      self._log('Engine initialized...')

# ------------------------------- Class Methods ------------------------------ #

  def _log(self, msg: str, end: str = None):
    """Print message to output if not silent

    Args:
        msg (str): Message to print to output
        end (str): String appended after the last value. Default a newline.
    """
    if self._verbose:
        print(msg, end=end)

  def _find_entry(self):
    """Waits for strategy entry

    Raises:
      Exception: Failed to find entry
    """
    try:
      # Wait until strategy finds entry signal
      while not self._strategy.entry():
        sleep(self._rate)
    except Exception as e:
      print('Failed to find entry...')
      raise

  def _open_position(self):
    """Open a position

    Raises:
      Exception: Failed to place market order
      Exception: Failed to place limit order
      InvalidParamError: type can only be limit or market

    Returns
      OrderClient: Order object
    """
    type_of_trade = self._strategy.params['type']

    # Place either market or limit order
    if type_of_trade == 'market':
      try:
        self._log('Placing a market order...')
        return self._open_market_trade()
      except Exception as e:
        print('Failed to place market order')
        raise
    elif type_of_trade == 'limit':
      try:
        self._log('Placing a limit order...')
        return self._open_limit_trade()
      except Exception as e:
        print('Failed to place limit order')
        raise
    else:
      raise InvalidParamError('type can only be limit or market')

  def _open_market_trade(self):
    """Places a market order

    Raises:
      Exception: Failed to calculate tp/sl
      Exception: Failed to place long order
      Exception: Failed to place short order

    Returns
      OrderClient: Order object
    """
    type_of_trade = self._strategy.params['type']
    side = self._strategy.params['side']
    amount = self._strategy.params['amount']
    tp_percent = self._strategy.params['tp_percent']
    sl_percent = self._strategy.params['sl_percent']
    symbol = self._strategy.params['symbol']

    try:
      # Calculate tp and sl
      self._log('Calculating take profit and stop loss prices...')
      price = self._proxy.price(symbol)
      tp = take_profit(side, price, tp_percent)
      sl = stop_loss(side, price, sl_percent)
      self._log(f'\nTake Profit: {tp}\nStop Loss: {sl}\n')
    except Exception as e:
      print('Failed to calculate tp/sl')
      raise
    finally:
      self._position_stats['entry'] = price
      self._position_stats['tp'] = tp
      self._position_stats['sl'] = sl

    # Place order
    if side == 'long':
      try:
        self._log('Opening a long position...')
        return self._proxy.long(symbol, type_of_trade, amount, sl=sl, tp=tp)
      except Exception as e:
        print('Failed to place long order')
        raise
    elif side == 'short':
      try:
        self._log('Opening a short position...')
        return self._proxy.short(symbol, type_of_trade, amount, sl=sl, tp=tp)
      except Exception as e:
        print('Failed to place short order')
        raise
    else:
      raise InvalidParamError('side can only be long or short')

  # TODO
  def _open_limit_trade(self):
    pass

  def _manage_position(self):
    """Manage open position

    Raises
      Exception: Failed to close position
      Exception: Failed to find exit strategy
      InvalidParamError: exit must be either True or False
    """
    exit_strategy = self._strategy.params['exit']
    # Check if there is an exit strategy already in place
    # If not then waits for tp/sl to get hit
    if not exit_strategy:
      try:
        self._close_position()
      except Exception as e:
        print('Failed to close position')
        raise
    elif exit_strategy:
      try:
        self._find_exit()
      except Exception as e:
        print('Failed to find exit strategy')
        raise
    else:
      raise InvalidParamError('exit must be either True or False')

  def _close_position(self):
    """Waits for position to be closed

    Raises:
      Exception: Failed to retrieve position
      Exception: Failed to close position
    """
    symbol = self._strategy.params['symbol']

    try:
      # Get open position
      self._log('Retrieving position...')
      position = self._proxy.position(symbol)
      self._log(f'Position found: \n{position}')
    except Exception as e:
      print('Failed to retrieve position')
      raise

    try:
      # Wait for position to be closed
      self._log('Waiting for position to be closed...')
      while not position._check_closed():
        self.view_position_stats()
        sleep(self._rate)
    except Exception as e:
      print('Failed to check if position was closed')
      raise

  # TODO
  def _find_exit(self):
    pass

  def _update_data(self):
    """Update trade stats and wallet

    Raises:
      Exception - Failed to update wallet
    """
    self._trade_stats['trade'] += 1
    prev_balance = self._wallet['balance']

    try:
      self._wallet['balance'] = self._proxy.balance(currency='USD', code='future')
    except Exception as e:
      print('Failed to update wallet')
      raise
    finally:
      # Win
      if prev_balance < self._wallet['balance']:
        self._wallet['pnl'] += self._wallet['balance'] - prev_balance
        self._trade_stats['wins'] += 1

      # Loss
      if prev_balance > self._wallet['balance']:
        self._wallet['pnl'] += prev_balance - self._wallet['balance']
        self._trade_stats['losses'] += 1

# ------------------------------ Client Methods ------------------------------ #

  def verbose(self):
      """Turn on logging"""
      self._verbose = True

  def silent(self):
      """Turn off logging"""
      self._verbose = False

  def view_trade_stats(self):
    """Output stats"""
    print(f'\nTrade: {self._trade_stats["trade"]}\nWins: {self._trade_stats["wins"]}\nLosses: {self._trade_stats["losses"]}\n')

  def view_position_stats(self):
    """Output position stats

    Raises:
      Exception - Failed to retrieve price and view position stats
    """
    try:
      print(f'\nPrice: {self._proxy(self._strategy.params["symbol"])}\nEntry: ${self._position_stats["entry"]}\nTake Profit: ${self._position_stats["tp"]}\nStop Loss: ${self._position_stats["sl"]}\n')
    except Exception as e:
      print('Failed to retrieve price and view position stats')
      raise

  def view_wallet(self):
    """Output wallet information"""
    print(f'\nBalance: ${self._wallet["balance"]}\nProfit/Loss: {self._wallet["pnl"]} (USD)\n')

  def run(self, trades: int = 1):
    """Run the strategy

    Args:
        trades (int, optional): Total number of trades to run. Defaults to 1.
    """

    while self._trade_stats['trade'] <= trades:
      try:
        # Wait until next minute
        self._log('Syncing...')
        sync()
        # Find entry
        self.view_trade_stats()
        self.view_wallet()
        self._log('Finding entry...')
        self._find_entry()
        # Open position
        self._log('Opening a position...')
        order = self._open_position()
        self._log(f'Order found: \{order}')
        # Check if order was closed
        self._log('Checking if order was closed...')
        if order.closed():
          self._log(f'Order closed, position opened...')
          # Mange open position
          self._manage_position()
        # Trade finished
        # Update trade data
        self._log('Trade complete, updating trade data...')
        self._update_data()
      except InvalidParamError as e:
        print(f'Strategy contains invalid params: \n{e}')
      except Exception as e:
        print(f'Error while running strategy: \n{e}')
        break
        # TODO: Cleanup method



"""Calculations helper"""

def take_profit(side: str, price: float, percent: float):
  """Calculate take profit price

  Args:
      side (str): Trade side (ex. 'long' or 'short')
      price (float): Price to calculate take profit for
      percent (float): Take profit percent

  Returns:
      Float: Take profit price
  """
  if side == 'long': return price + (price * (percent / 100))
  if side == 'short': return price - (price * (percent / 100))

def stop_loss(side: str, price: float, percent: float):
  """Calculate stop loss price

  Args:
      side (str): Trade side (ex. 'long' or 'short')
      price (float): Price to calculate stop loss for
      percent (float): Stop loss percent

  Returns:
      Float: Stop loss price
  """
  if side == 'long': return price - (price * (percent / 100))
  if side == 'short': return price + (price * (percent / 100))



"""Runs example strategy"""

from tradeboy.engines.future import FutureEngine
from strategies.example import Strategy as ex_strat

engine = FutureEngine(Strategy=ex_strat, rate=30, verbose=True)
engine.run(trades=100)



"""Strategy helpers"""

from phemexboy.proxy import Proxy
from candleboy.core import CandleBoy

# TODO: Error handling
# TODO: Tests

class Tools:
  def __init__(self, exchange: str, verbose: bool = False):
    self.proxy = Proxy(verbose)
    self.candle = CandleBoy(exchange, verbose)

  def symbol(self, base: str, quote: str, code: str):
    """Creates a symbol representing the asset pairing

    Args:
        base (str): Currency you are buying (ex. 'btc')
        quote (str): Currency you are selling (ex. 'usd')
        code (str): Market code (ex. 'spot')

    Returns:
        String: Formatted base and quote currency symbol
    """
    return self.proxy.symbol(base, quote, code)

  def price(self, symbol: str):
    """Retrieve price of asset pair

    Args:
        symbol (str): Created symbol for base and quote currencies

    Raises:
        NetworkError: PublicClient failed to retrieve price for {symbol}
        ExchangeError: PublicClient failed to retrieve price for {symbol}
        Exception: PublicClient failed to retrieve price for {symbol}

    Returns:
        Float: Current ask price for base currency
    """
    return self.proxy.price(symbol)

  def ohlcv(self, symbol: str, tf: str, since: str = None):
    """Retrieve the open - high - low - close - volume data from exchange

    Args:
        symbol (str): Pairing to retrieve OHLCV data for
        tf (str): Timeframe for OHLCV data
        since (str, optional): Optional start date for retrieving OHLCV data. Format as YEAR-MONTH-DAY (ex. 2018-12-01). Defaults to None.

    Returns:
        Tuple: open - high - low - close - volume data
    """
    return self.candle.ohlcv(symbol, tf, since)

  def macd(self, close: list, fastperiod: int = 12, slowperiod: int = 26, signalperiod: int = 9):
    """Returns the Moving Average Convergence/Divergence indicator values

    See TA-Lib docs for details on parameters.
    https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
    """
    return self.candle.macd(close, fastperiod, slowperiod, signalperiod)

  def ema(self, close: list, timeperiod: int = 200):
    """Returns the Exponential Moving Average indicator values

    See TA-Lib docs for details on parameters.
    https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
    """
    return self.candle.ema(close, timeperiod)

  def stoch(self, high: list, low: list, close: list, fastk_period: int = 5, slowk_period: int = 3, slowk_matype: int = 0, slowd_period: int = 3, slowd_matype: int = 0):
    """Returns the Stochastic indicator values

    See TA-Lib docs for details on parameters.
    https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
    """
    return self.candle.stoch(high, low, close, fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype)
